###############################################################################
#                                                                             #
#          Copyright (C) 2021 LaVision GmbH.  All Rights Reserved.            #
#                                                                             #
###############################################################################

import sys
from os import path
from glob import glob
from shutil import copyfile
from tqdm import tqdm
import lvpyio as lv
import cv2


def resize(in_set, *, xy_only):

    out_set = []
    count = len(in_set)
    progress = tqdm(enumerate(in_set, 1), total=count, bar_format="{l_bar}{bar}", leave=False)

    for i, buffer in progress:
        progress.set_description(f"Resizing image {i} of {count}")
        frame = buffer[0]  # multi-frame data not expected

        for scale in frame.scales.x, frame.scales.y, frame.scales.z:
            scale.slope *= 2

        # A DVC recording usually has one component (PIXEL). This is what we want to resize.
        # But lvpyio output, for simplicity, always has MASK, even if it's all ones.
        # Resizing the PIXEL component and keeping the old MASK would lead to errors.
        # To allow the user to do recursive resizing, we delete the mask:
        if "MASK" in frame.components:
            del frame.components["MASK"]

        images = frame.components["PIXEL"].planes
        in_image = images[0]  # multi-plane data not expected
        out_image = cv2.resize(in_image, None, fx=0.5, fy=0.5, interpolation=cv2.INTER_AREA)
        images[0] = out_image

        # When resizing along x and y, but not z:
        if xy_only:
            out_set.append(buffer)
            continue

        # When resizing along x, y and z:
        if i == count:
            # Ignore the last image if the number of images is odd.
            # This is consistent with behavior of OpenCV resize.
            pass
        elif i % 2 == 1:
            previous_image = out_image
            out_set.append(buffer)
        else:
            previous_image += out_image
            previous_image //= 2

    return out_set


def main():

    if len(sys.argv) == 3:
        in_path, out_path = sys.argv[1:]
    else:
        print("Usage:\n  python dvc_resize.py <input path> <output path>")
        sys.exit(1)

    in_data_paths = glob(path.join(in_path, "Data", "S*.set"))
    assert in_data_paths, "'Data' folder not found"

    in_set = lv.read_set(in_path)
    print("Root set")
    out_set = resize(in_set, xy_only=True)
    lv.write_set(out_set, out_path)

    copyfile(path.join(in_path, "Data.set"), path.join(out_path, "Data.set"))

    count = len(in_data_paths)

    for i, in_data_path in enumerate(in_data_paths, 1):
        print(f"Volume {i} of {count}")

        _, filename = path.split(in_data_path)
        out_data_path = path.join(out_path, "Data", filename)

        in_set = lv.read_set(in_data_path)
        out_set = resize(in_set, xy_only=False)

        lv.write_set(out_set, out_data_path)


if __name__ == "__main__":
    main()
