###############################################################################
#                                                                             #
#             Copyright (C) LaVision GmbH.  All Rights Reserved.              #
#                                                                             #
###############################################################################

import os
import sys
import argparse
import warnings
import h5py
import numpy as np
import lvpyio as lv


def calculate_displacements(array):
    height, width = array.shape
    is_3c = len(array.dtype) == 3

    vector_components = ["u", "v", "w"] if is_3c else ["u", "v"]

    displacements = np.empty((3 if is_3c else 2, height, width), dtype=np.float64)
    for i, key in enumerate(vector_components):
        displacements[i] = array[key]

    return displacements


def calculate_coordinates(frame):
    # Height is the y-dimension.
    # Not to be confused with the Height scalar field stored in TS:Height.
    height, width = frame.shape

    if frame.is_3c:
        scales = [frame.scales.x, frame.scales.y, frame.scales.z]
    else:
        scales = [frame.scales.x, frame.scales.y]

    coordinates = np.empty((3 if frame.is_3c else 2, height, width), dtype=np.float64)

    coordinates_x = frame.grid.x * np.arange(width) + frame.grid.x // 2
    for i in range(height):
        coordinates[0, i, :] = coordinates_x
    coordinates[0] = scales[0].slope * coordinates[0] + scales[0].offset

    coordinates_y = frame.grid.y * np.arange(height) + frame.grid.y // 2
    for i in range(width):
        coordinates[1, :, i] = coordinates_y
    coordinates[1] = scales[1].slope * coordinates[1] + scales[1].offset

    if frame.is_3c:
        assert "TS:Height" in frame.components, "3C Strain data must have 'Height' scalar field"
        height_scale = frame.components["TS:Height"].scale
        # Z component scale should be equal to height scale.
        # Only their descriptions are allowed to be different:
        scales_match = (
            height_scale.slope == scales[2].slope and
            height_scale.offset == scales[2].offset and
            height_scale.unit == scales[2].unit
        )

        if not scales_match:
            warnings.warn("z component and 'Height' scalar field have different scales", RuntimeWarning)
        coordinates[2] = scales[2].slope * frame[0]["TS:Height"] + scales[2].offset

    units = [scale.unit for scale in scales]
    descriptions = [scale.description for scale in scales]
    return coordinates, units, descriptions


def fill_buffer_group(buffer_group, buffer):
    if len(buffer) != 1:
        warnings.warn(f"expected single-frame data, got {len(buffer)} frames", RuntimeWarning)
    frame = buffer[0]

    if len(frame) != 1:
        warnings.warn(f"expected 2D data, got {len(frame)} planes", RuntimeWarning)

    if frame.type_id != 'VectorFrame':
        raise RuntimeError(f"expected VectorFrame, got {frame.type_id}")

    scaled_masked_array = frame.as_masked_array()

    displacements = calculate_displacements(scaled_masked_array)
    displacement_unit = frame.scales.i.unit
    displacement_description = frame.scales.i.description
    coordinates, coordinate_units, coordinate_descriptions = calculate_coordinates(frame)
    # position in hdf5 (actual) = position in vc7 (reference) + deformation
    coordinates += displacements

    mask = scaled_masked_array.mask["u"].astype(np.uint8)

    timestamp_key = "Timestamp"
    if timestamp_key in buffer.attributes:
        buffer_group.attrs[timestamp_key] = buffer.attributes[timestamp_key]
    else:
        warnings.warn(f"buffer attribute {timestamp_key} not found", RuntimeWarning)

    exposure_time_key = "CCDExposureTime"
    if exposure_time_key in frame.attributes:
        buffer_group.attrs[exposure_time_key] = frame.attributes[exposure_time_key]
    else:
        warnings.warn(f"frame attribute {exposure_time_key} not found", RuntimeWarning)

    coordinates_group = buffer_group.create_group("coordinates")
    coordinates_group.attrs["Unit"] = coordinate_units
    coordinates_group.attrs["Description"] = coordinate_descriptions

    displacements_group = buffer_group.create_group("displacements")
    displacements_group.attrs["Unit"] = displacement_unit
    displacements_group.attrs["Description"] = displacement_description

    coordinates_group.create_dataset("coordinate_x", data=coordinates[0], compression="gzip")  # x
    coordinates_group.create_dataset("coordinate_y", data=coordinates[1], compression="gzip")  # y
    displacements_group.create_dataset("displacement_x", data=displacements[0], compression="gzip")  # dx
    displacements_group.create_dataset("displacement_y", data=displacements[1], compression="gzip")  # dy

    if len(coordinates) == 3:
        coordinates_group.create_dataset("coordinate_z", data=coordinates[2], compression="gzip")  # z
        displacements_group.create_dataset("displacement_z", data=displacements[2], compression="gzip")  # dz

    total_displacement = np.sqrt(np.sum(displacements**2, axis=0))
    displacements_group.create_dataset("displacement_total", data=total_displacement, compression="gzip")

    mask_type = h5py.enum_dtype({"VALID": 0, "INVALID": 1}, basetype=np.uint8)
    coordinates_group.create_dataset("mask", data=mask, compression="gzip", dtype=mask_type)
    # hard link, see https://docs.h5py.org/en/latest/high/group.html#hard-links
    displacements_group["mask"] = coordinates_group["mask"]


def create_buffer_group(hdf5_file, buffer):
    """
    Create and fill a group in the hdf5 file. The group corresponds to one vector buffer.
    """
    path = buffer.attributes["LoadFile"]
    dir_name, file_name = os.path.split(path)
    group_name, _ = os.path.splitext(file_name)

    group = hdf5_file.create_group(group_name)
    group.attrs["VC7 Path"] = dir_name
    group.attrs["VC7 File"] = file_name
    fill_buffer_group(group, buffer)
    return group


def read_buffer_write_hdf5(input_buffer_path, hdf5_file_path):
    buffer = lv.read_buffer(input_buffer_path)
    print(f"Writing {hdf5_file_path}...")
    with h5py.File(hdf5_file_path, "w") as file:
        create_buffer_group(file, buffer)


def read_set_write_hdf5(input_set_path, hdf5_file_path):
    input_set = lv.read_set(input_set_path)
    print(f"Writing {hdf5_file_path}...")
    with h5py.File(hdf5_file_path, "w") as file:
        for buffer in input_set:
            group = create_buffer_group(file, buffer)
            group.attrs["Step No."] = int(buffer.attributes["LoadSetIndex"])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('input_path', help='path to input set (.set) or input vector buffer (.vc7)')
    parser.add_argument('output_path', nargs='?', help='path to output file (.hdf5)')
    parser.add_argument('--overwrite', dest="overwrite", action="store_true", help='overwrite output file if it exists')
    args = parser.parse_args()

    input_path = os.path.normpath(args.input_path)
    root, ext = os.path.splitext(input_path)

    if args.output_path:
        hdf5_file_path = os.path.normpath(args.output_path)
    else:
        hdf5_file_path = f"{root}.hdf5"
        print(f"No output file path specified, assuming {hdf5_file_path}")

    if not args.overwrite and os.path.exists(hdf5_file_path):
        sys.stderr.write(f"Error: {hdf5_file_path} already exists, specify `--overwrite` to overwrite it")
        sys.exit(1)

    if ext.lower() in ("", ".set"):
        read_set_write_hdf5(input_path, hdf5_file_path)
    elif ext.lower() in ("vec", ".vc7"):
        read_buffer_write_hdf5(input_path, hdf5_file_path)
    else:
        raise RuntimeError(f"can't convert {ext} to hdf5")


if __name__ == "__main__":
    main()
